import os
from logging import Logger
from multiprocessing.pool import ThreadPool
from typing import Dict, List, Type, Optional

from botocore.client import BaseClient
from lambda_thread_pool import LambdaThreadPool

from ..types import (
    FinalSpotPriceStrategy,
    InstanceTypeInfo,
    ProductDescription,
    PriceDetails,
    TypePriceDetails,
)
from ..aws_utils import AwsUtils
from ..exceptions import InvalidStrategyError
from .abstract_sort_strategy import AbstractSortStrategy


class SortSpotStrategy(AbstractSortStrategy):
    def __init__(
        self,
        region: str,
        pricing_client: BaseClient,
        ec2_client: BaseClient,
        logger: Logger,
        spot_price_history_concurrency: int,
    ):
        super().__init__(
            region, pricing_client, ec2_client, logger, spot_price_history_concurrency
        )
        self._pool: Type[ThreadPool] = (
            LambdaThreadPool
            if os.environ.get("AWS_LAMBDA_FUNCTION_NAME")
            else ThreadPool
        )
        self._aws_utils = AwsUtils(ec2_client)

    def _get_price(
        self,
        filtered_instances: List[InstanceTypeInfo],
        product_description: ProductDescription,
        availability_zones: Optional[List[str]],
        final_spot_price_strategy: FinalSpotPriceStrategy,
    ) -> Dict[str, PriceDetails]:
        availability_zones = (
            availability_zones
            or self._aws_utils.get_all_availability_zones_for_region()
        )

        with self._pool() as executor:
            futures = [
                executor.apply_async(
                    self._get_spot_instance_price,
                    (
                        ec2_instance,
                        product_description,
                        availability_zones,
                        final_spot_price_strategy,
                    ),
                )
                for ec2_instance in filtered_instances
            ]

            ec2_prices = {
                f.get()["instance_type"]: f.get()["price_details"]
                for f in futures
                if f.get()
            }

        return ec2_prices

    def _get_spot_instance_price(
        self,
        ec2_instance: InstanceTypeInfo,
        product_description: ProductDescription,
        availability_zones: List[str],
        strategy: FinalSpotPriceStrategy,
    ) -> Optional[TypePriceDetails]:
        instance_type = ec2_instance["InstanceType"]
        filters = [{"Name": "product-description", "Values": [product_description]}]

        if availability_zones:
            filters.append({"Name": "availability-zone", "Values": availability_zones})

        try:
            response = self._ec2_client.describe_spot_price_history(
                InstanceTypes=[instance_type], Filters=filters
            )
        except Exception as e:
            self._logger.error(f"Error fetching price history: {e}")
            return None

        history_events = response["SpotPriceHistory"]
        az_price: Dict[str, float] = {
            event["AvailabilityZone"]: float(event["SpotPrice"])
            for event in history_events
            if event["AvailabilityZone"] in availability_zones
        }

        spot_price = self._calculate_spot_price(list(az_price.values()), strategy)

        return {
            "instance_type": instance_type,
            "price_details": {"price": spot_price, "az_price": az_price},
        }

    @staticmethod
    def _calculate_spot_price(
        prices: List[float], strategy: FinalSpotPriceStrategy
    ) -> float:
        if strategy == "average":
            return sum(prices) / len(prices)
        elif strategy == "max":
            return max(prices)
        elif strategy == "min":
            return min(prices)
        else:
            raise InvalidStrategyError(strategy)



from best_ec2 import (
    BestEc2,
    InstanceTypeRequest,
    InstanceTypeResponse,
    UsageClass,
    BestEc2Options,
    Architecture,
    ProductDescription,
    FinalSpotPriceStrategy,
)


def test_on_demand_instance_types_lack_az_price():
    ec2 = BestEc2()

    request: InstanceTypeRequest = {"vcpu": 1, "memory_gb": 1}

    response: InstanceTypeResponse = ec2.get_types(request)

    assert response is not None, "Response should not be None"

    for instance_type in response:
        assert (
            "az_price" not in instance_type
        ), f"Instance type {instance_type.get('instance_type', 'unknown')} should not have 'az_price' key."


def test_spot_instance_types_have_az_price():
    ec2 = BestEc2()

    request: InstanceTypeRequest = {
        "vcpu": 1,
        "memory_gb": 1,
        "usage_class": UsageClass.SPOT.value,
    }

    response: InstanceTypeResponse = ec2.get_types(request)

    assert response is not None, "Response should not be None"

    for instance_type in response:
        assert (
            "az_price" in instance_type
        ), f"Instance type {instance_type.get('instance_type', 'unknown')} must have 'az_price' key."


def test_on_demand_gpu():
    ec2 = BestEc2()

    request: InstanceTypeRequest = {
        "vcpu": 1,
        "memory_gb": 1,
        "usage_class": UsageClass.ON_DEMAND.value,
        "has_gpu": True,
    }

    response: InstanceTypeResponse = ec2.get_types(request)

    for instance_type in response:
        assert "gpu_memory_gb" in instance_type

    assert response is not None, "Response should not be None"


def test_advanced():
    from botocore.config import Config
    import logging
    import boto3

    ec2_client_config = Config(retries={"max_attempts": 20, "mode": "adaptive"})

    pricing_client_config = Config(retries={"max_attempts": 10, "mode": "standard"})

    ec2_client = boto3.Session().client(
        "ec2", config=ec2_client_config, region_name="eu-central-1"
    )
    pricing_client = boto3.Session().client(
        "pricing", config=pricing_client_config, region_name="us-east-1"
    )

    options: BestEc2Options = {
        "region": "eu-central-1",
        "describe_spot_price_history_concurrency": 20,
        "describe_on_demand_price_concurrency": 20,
        "clients": {"ec2": ec2_client, "pricing": pricing_client},
        "cache_ttl_in_minutes": 60,
    }

    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s: %(levelname)s: %(message)s"
    )
    logger = logging.getLogger()

    ec2 = BestEc2(options, logger)

    request: InstanceTypeRequest = {
        "vcpu": 1,
        "memory_gb": 2,
        "usage_class": UsageClass.SPOT.value,
        "burstable": False,
        "architecture": Architecture.X86_64.value,
        "product_description": ProductDescription.LINUX_UNIX.value,
        "is_current_generation": True,
        "is_instance_storage_supported": True,
        "max_interruption_frequency": 10,
        "availability_zones": [
            "eu-central-1a",
            "eu-central-1b",
        ],
        "final_spot_price_strategy": FinalSpotPriceStrategy.MIN.value,
    }

    response: InstanceTypeResponse = ec2.get_types(request)

    assert response is not None, "Response should not be None"



from abc import ABC, abstractmethod
from typing import List

from .constants import OS_PRODUCT_DESCRIPTION_MAP
from .spot_utils import SpotUtils
from .types import InstanceTypeInfo, InstanceTypeRequest, UsageClass


class Filter(ABC):
    @abstractmethod
    def apply(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        pass


class VcpuFilter(Filter):
    def apply(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        return [
            i
            for i in instances
            if i["VCpuInfo"]["DefaultVCpus"] >= request.get("vcpu", 0)
        ]


class MemoryFilter(Filter):
    def apply(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        return [
            i
            for i in instances
            if i["MemoryInfo"]["SizeInMiB"] >= request.get("memory_gb", 0) * 1024
        ]


class UsageClassFilter(Filter):
    def apply(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        usage_class = request.get("usage_class")
        if usage_class is None:
            return instances
        return [i for i in instances if usage_class in i["SupportedUsageClasses"]]


class BurstableFilter(Filter):
    def apply(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        burstable = request.get("burstable")
        return (
            [i for i in instances if i["BurstablePerformanceSupported"] == burstable]
            if burstable is not None
            else instances
        )


class ArchitectureFilter(Filter):
    def apply(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        architecture = request.get("architecture")
        return (
            [
                i
                for i in instances
                if architecture in i["ProcessorInfo"]["SupportedArchitectures"]
            ]
            if architecture is not None
            else instances
        )


class GpuFilter(Filter):
    def apply(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        has_gpu = request.get("has_gpu")
        gpu_memory = request.get("gpu_memory", 0)

        if has_gpu is None:
            return instances

        if has_gpu:
            return [
                i
                for i in instances
                if "GpuInfo" in i
                and i["GpuInfo"]["TotalGpuMemoryInMiB"] >= gpu_memory * 1024
            ]
        return [i for i in instances if "GpuInfo" not in i]


class SpotFilter(Filter):
    def __init__(self, region: str):
        self._region = region

    def apply(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        if (
            request.get("usage_class") != UsageClass.SPOT.value
            or "max_interruption_frequency" not in request
        ):
            return instances

        spot_utils = SpotUtils(self._region)
        operating_system = OS_PRODUCT_DESCRIPTION_MAP.get(
            request["product_description"]
        )
        interruption_frequencies = spot_utils.get_spot_interruption_frequency(
            operating_system
        )

        for instance in instances:
            instance_type = instance["InstanceType"]
            instance["InterruptionFrequency"] = interruption_frequencies[instance_type]

        max_freq = request["max_interruption_frequency"]

        return [
            instance
            for instance in instances
            if instance["InterruptionFrequency"]["min"] <= max_freq
        ]


class FilterChain:
    def __init__(self, region: str):
        self._filters: List[Filter] = [
            VcpuFilter(),
            MemoryFilter(),
            UsageClassFilter(),
            BurstableFilter(),
            ArchitectureFilter(),
            GpuFilter(),
            SpotFilter(region),
        ]

    def execute(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        for filter_ in self._filters:
            instances = filter_.apply(instances, request)
        return instances



import json
from urllib import request, error

SPOT_ADVISOR_JSON_URL = (
    "https://spot-bid-advisor.s3.amazonaws.com/spot-advisor-data.json"
)


class SpotUtils:
    def __init__(self, region):
        self.__region = region

    def get_spot_interruption_frequency(self, os):
        results = {}
        with request.urlopen(SPOT_ADVISOR_JSON_URL) as response:
            response_data = response.read().decode("utf-8")
            spot_advisor = json.loads(response_data)["spot_advisor"]

        rates = {
            0: {"min": 0, "max": 5, "rate": "<5%"},
            1: {"min": 6, "max": 10, "rate": "5-10%"},
            2: {"min": 11, "max": 15, "rate": "10-15%"},
            3: {"min": 16, "max": 20, "rate": "15-20%"},
            4: {"min": 21, "max": 100, "rate": ">20%"},
        }

        instance_data = spot_advisor[self.__region].get(os, {})

        for instance_type, data in instance_data.items():
            try:
                rate = data["r"]
                results[instance_type] = rates[rate]
            except (IndexError, KeyError):
                results[instance_type] = ""

        return results



from typing import List, Union

from botocore.client import BaseClient

from .exceptions import NoAvailabilityZonesError


class AwsUtils:
    def __init__(self, ec2_client: BaseClient):
        self._availability_zones_cache: Union[None, List[str]] = None
        self._ec2_client = ec2_client

    def get_region(self) -> str:
        return self._ec2_client.meta.region_name

    def get_all_availability_zones_for_region(self) -> List[str]:
        if self._availability_zones_cache is None:
            self._fetch_and_cache_availability_zones()

        return self._availability_zones_cache

    def _fetch_and_cache_availability_zones(self):
        response = self._ec2_client.describe_availability_zones()
        if availability_zones := [
            zone["ZoneName"] for zone in response.get("AvailabilityZones", [])
        ]:
            self._availability_zones_cache = availability_zones
        else:
            raise NoAvailabilityZonesError(self.get_region())



from abc import ABC, abstractmethod
import operator
from logging import Logger
from typing import Dict, List, Optional, Union

from botocore.client import BaseClient

from ..types import (
    FinalSpotPriceStrategy,
    InstanceTypeInfo,
    InstanceTypeResponse,
    ProductDescription,
    InstanceType,
    PriceDetails,
)


class AbstractSortStrategy(ABC):
    def __init__(
        self,
        region: str,
        pricing_client: BaseClient,
        ec2_client: BaseClient,
        logger: Logger,
        concurrency_level: int = 10,
    ):
        self._logger = logger
        self._region = region
        self._pricing_client = pricing_client
        self._ec2_client = ec2_client
        self._concurrency_level = concurrency_level

    def sort(
        self,
        instance_types_info: List[InstanceTypeInfo],
        product_desc: ProductDescription,
        availability_zones: List[str],
        pricing_strategy: Optional[FinalSpotPriceStrategy],
    ) -> InstanceTypeResponse:
        price_details_by_instance = self._get_price(
            instance_types_info,
            product_desc,
            availability_zones,
            pricing_strategy,
        )

        sorted_instances = []
        for instance_info in instance_types_info:
            instance_entry = self._build_instance_entry(
                instance_info, price_details_by_instance
            )
            if instance_entry is not None:
                sorted_instances.append(instance_entry)

        sorted_instances.sort(key=operator.itemgetter("price"))

        return sorted_instances

    @abstractmethod
    def _get_price(
        self,
        filtered_instances: List[InstanceTypeInfo],
        product_description: ProductDescription,
        availability_zones: Optional[List[str]],
        final_spot_price_strategy: FinalSpotPriceStrategy,
    ) -> Dict[str, PriceDetails]:
        pass

    def _build_instance_entry(
        self,
        instance_info: InstanceTypeInfo,
        prices: Dict[str, PriceDetails],
    ) -> Union[InstanceType, None]:
        instance_type_name = instance_info["InstanceType"]
        price_info = prices.get(instance_type_name)

        if price_info is None:
            self._logger.info(
                f"The price for the {instance_type_name} instance type not found"
            )
            return None

        entry: InstanceType = {
            "instance_type": instance_type_name,
            "vcpu": instance_info["VCpuInfo"]["DefaultVCpus"],
            "memory_gb": self._transform_memory_size(
                instance_info["MemoryInfo"]["SizeInMiB"]
            ),
            "network_performance": instance_info["NetworkInfo"]["NetworkPerformance"],
            "storage": instance_info.get("InstanceStorageInfo", {}).get(
                "Disks", "EBS Only"
            ),
            "price": price_info["price"],
        }

        if az_price := price_info.get("az_price"):
            entry["az_price"] = az_price

        if gpu_info := instance_info.get("GpuInfo"):
            entry["gpu_memory_gb"] = self._transform_memory_size(
                gpu_info["TotalGpuMemoryInMiB"]
            )

        if interruption_frequency := instance_info.get("InterruptionFrequency"):
            entry["interruption_frequency"] = interruption_frequency

        return entry

    @staticmethod
    def _transform_memory_size(memory_in_mib: int) -> int:
        """Converts memory size from MiB to GB"""
        return memory_in_mib // 1024



from typing import Optional
from logging import Logger

from .types import BestEc2Options, InstanceTypeRequest, InstanceTypeResponse
from .best_ec2_impl import BestEc2Impl


class BestEc2:
    def __init__(
        self, options: Optional[BestEc2Options] = None, logger: Optional[Logger] = None
    ):
        self.__impl: BestEc2Impl = BestEc2Impl(options, logger)

    def get_types(
        self, request: Optional[InstanceTypeRequest] = None
    ) -> InstanceTypeResponse:
        return self.__impl.get_types(request)



class NoAvailabilityZonesError(Exception):
    """Exception raised when no availability zones are found for a given AWS region."""

    def __init__(
        self, region, message="Availability zones not found for the specified region"
    ):
        self.region = region
        self.message = f"{message}: {region}"
        super().__init__(self.message)


class InvalidStrategyError(ValueError):
    """Exception raised when an invalid strategy is provided."""

    def __init__(self, strategy, message="The specified strategy is invalid"):
        self.strategy = strategy
        self.message = f"{message}: {strategy}"
        super().__init__(self.message)



import json
from typing import Dict, List, Optional, Tuple

from ..constants import OS_PRODUCT_DESCRIPTION_MAP, REGIONS
from ..types import (
    FinalSpotPriceStrategy,
    InstanceTypeInfo,
    ProductDescription,
    PriceDetails,
)
from .abstract_sort_strategy import AbstractSortStrategy


class SortOnDemandStrategy(AbstractSortStrategy):
    """Sort strategy for On-Demand EC2 instances."""

    def _get_price(
        self,
        filtered_instances: List[InstanceTypeInfo],
        product_description: ProductDescription,
        availability_zones: Optional[List[str]],
        final_spot_price_strategy: FinalSpotPriceStrategy,
    ) -> Dict[str, PriceDetails]:
        operating_system = self._map_product_description_to_os(product_description)
        return self._get_on_demand_instance_price(operating_system)

    @staticmethod
    def _map_product_description_to_os(product_description: ProductDescription) -> str:
        return OS_PRODUCT_DESCRIPTION_MAP[product_description]

    def _get_on_demand_instance_price(
        self, operating_system: str
    ) -> Dict[str, PriceDetails]:
        filters = self._build_ec2_filter_criteria(operating_system)
        return self._retrieve_pricing_records(filters)

    def _build_ec2_filter_criteria(self, operating_system: str) -> List[Dict[str, str]]:
        return [
            {"Type": "TERM_MATCH", "Field": "preInstalledSw", "Value": "NA"},
            {
                "Type": "TERM_MATCH",
                "Field": "productFamily",
                "Value": "Compute Instance",
            },
            {"Type": "TERM_MATCH", "Field": "termType", "Value": "OnDemand"},
            {"Type": "TERM_MATCH", "Field": "location", "Value": REGIONS[self._region]},
            {
                "Type": "TERM_MATCH",
                "Field": "licenseModel",
                "Value": "No License required",
            },
            {"Type": "TERM_MATCH", "Field": "tenancy", "Value": "Shared"},
            {"Type": "TERM_MATCH", "Field": "capacitystatus", "Value": "Used"},
            {
                "Type": "TERM_MATCH",
                "Field": "operatingSystem",
                "Value": operating_system,
            },
        ]

    def _retrieve_pricing_records(
        self, filters: List[Dict[str, str]]
    ) -> Dict[str, PriceDetails]:
        records = {}
        next_token = None

        while True:
            response = self._query_pricing_api(filters, next_token)
            price_list = response.get("PriceList", [])

            if not price_list:
                break

            for price in price_list:
                instance_type, instance_price = self._parse_price_details(price)
                if instance_price > 0:
                    records[instance_type] = {"price": instance_price}

            next_token = response.get("NextToken")
            if not next_token:
                break

        return records

    def _query_pricing_api(
        self, filters: List[Dict[str, str]], next_token: Optional[str]
    ) -> dict:
        request_parameters = {
            "ServiceCode": "AmazonEC2",
            "Filters": filters,
        }

        if next_token:
            request_parameters["NextToken"] = next_token

        return self._pricing_client.get_products(**request_parameters)

    @staticmethod
    def _parse_price_details(price: str) -> Tuple[str, float]:
        details = json.loads(price)
        price_dimensions = next(iter(details["terms"]["OnDemand"].values()))[
            "priceDimensions"
        ]
        pricing_details = next(iter(price_dimensions.values()))
        instance_price = float(pricing_details["pricePerUnit"]["USD"])
        instance_type = details["product"]["attributes"]["instanceType"]

        return instance_type, instance_price



import hashlib
import json
from datetime import datetime, timedelta
from typing import Optional

from .types import (
    CacheDict,
    ExtendedInstanceTypeRequest,
    InstanceTypeRequest,
    InstanceTypeResponse,
)


class Cache:
    def __init__(self, ttl_minutes: int, region: str):
        self._region = region
        self._ttl_minutes = ttl_minutes
        self._cache: CacheDict = {}

    def get(self, request: InstanceTypeRequest) -> Optional[InstanceTypeResponse]:
        hash_digest = self._get_hash(request)

        cached = self._cache.get(hash_digest)
        if cached and self._is_valid(cached["datetime"]):
            return cached["result"]
        return None

    def set(self, request: InstanceTypeRequest, result: InstanceTypeResponse) -> None:
        hash_digest = self._get_hash(request)
        self._cache[hash_digest] = {"result": result, "datetime": datetime.now()}

    def _get_hash(self, request: InstanceTypeRequest) -> str:
        cached_request: ExtendedInstanceTypeRequest = dict(request)
        cached_request["region"] = self._region
        hash_object = hashlib.md5(json.dumps(cached_request, sort_keys=True).encode())
        return hash_object.hexdigest()

    def _is_valid(self, cache_datetime: datetime) -> bool:
        return (datetime.now() - cache_datetime) < timedelta(minutes=self._ttl_minutes)



from logging import Logger
from botocore.client import BaseClient
from ..types import UsageClass
from .abstract_sort_strategy import AbstractSortStrategy
from .sort_on_demand_strategy import SortOnDemandStrategy
from .sort_spot_strategy import SortSpotStrategy


class SortStrategyFactory:
    @staticmethod
    def create(
        usage_class: UsageClass,
        region: str,
        pricing_client: BaseClient,
        ec2_client: BaseClient,
        logger: Logger,
        concurrency: int,
    ) -> AbstractSortStrategy:
        strategy_map = {
            UsageClass.ON_DEMAND.value: SortOnDemandStrategy,
            UsageClass.SPOT.value: SortSpotStrategy,
        }

        strategy_class = strategy_map.get(usage_class)
        if strategy_class is None:
            raise ValueError(f"Unsupported usage class: {usage_class}")

        return strategy_class(region, pricing_client, ec2_client, logger, concurrency)



DEFAULT_CACHE_TTL_IN_MINUTES = 120  # Default time to live for cache entries, in minutes
DEFAULT_SPOT_CONCURRENCY = 10
DEFAULT_ON_DEMAND_CONCURRENCY = 10

# Default AWS region for operations
DEFAULT_REGION = "us-east-1"

REGIONS = {
    "us-east-2": "US East (Ohio)",
    "us-east-1": "US East (N. Virginia)",
    "us-west-1": "US West (N. California)",
    "us-west-2": "US West (Oregon)",
    "ap-east-1": "Asia Pacific (Hong Kong)",
    "ap-south-1": "Asia Pacific (Mumbai)",
    "ap-northeast-3": "Asia Pacific (Osaka-Local)",
    "ap-northeast-2": "Asia Pacific (Seoul)",
    "ap-southeast-1": "Asia Pacific (Singapore)",
    "ap-southeast-2": "Asia Pacific (Sydney)",
    "ap-northeast-1": "Asia Pacific (Tokyo)",
    "ca-central-1": "Canada (Central)",
    "cn-north-1": "China (Beijing)",
    "cn-northwest-1": "China (Ningxia)",
    "eu-central-1": "EU (Frankfurt)",
    "eu-west-1": "EU (Ireland)",
    "eu-west-2": "EU (London)",
    "eu-west-3": "EU (Paris)",
    "eu-north-1": "EU (Stockholm)",
    "me-south-1": "Middle East (Bahrain)",
    "sa-east-1": "South America (Sao Paulo)",
}

# Short names for EC2 operating system product descriptions
OS_PRODUCT_DESCRIPTION_MAP = {
    "Linux/UNIX": "Linux",
    "Red Hat Enterprise Linux": "RHEL",
    "SUSE Linux": "SUSE",
    "Windows": "Windows",
    "Linux/UNIX (Amazon VPC)": "Linux",
    "Red Hat Enterprise Linux (Amazon VPC)": "RHEL",
    "SUSE Linux (Amazon VPC)": "SUSE",
    "Windows (Amazon VPC)": "Windows",
}



from logging import Logger, getLogger, StreamHandler
from typing import Optional, List

import boto3
from botocore.client import BaseClient

from .types import (
    BestEc2Options,
    DescribeInstanceTypeRequest,
    InstanceTypeRequest,
    InstanceTypeResponse,
    ProductDescription,
    UsageClass,
    Architecture,
    InstanceTypeInfo,
)
from .constants import (
    DEFAULT_CACHE_TTL_IN_MINUTES,
    DEFAULT_SPOT_CONCURRENCY,
    DEFAULT_REGION,
)
from .strategies.sort_strategy_factory import SortStrategyFactory
from .cache import Cache
from .validators import Validator
from .filters import FilterChain


class BestEc2Impl:
    def __init__(
        self, options: Optional[BestEc2Options] = None, logger: Optional[Logger] = None
    ):
        options = options if options is not None else {}
        self._region = options.get("region", DEFAULT_REGION)
        cache_ttl = options.get("cache_ttl_in_minutes", DEFAULT_CACHE_TTL_IN_MINUTES)
        self._cache = Cache(cache_ttl, self._region)
        self._ec2_client = self._get_client(options, "ec2", self._region)
        self._pricing_client = self._get_client(options, "pricing", "us-east-1")
        self._logger = logger or self._setup_default_logger()
        self._spot_price_history_concurrency = options.get(
            "describe_spot_price_history_concurrency", DEFAULT_SPOT_CONCURRENCY
        )

    @staticmethod
    def _setup_default_logger() -> Logger:
        logger = getLogger(__name__)
        console_handler = StreamHandler()
        logger.addHandler(console_handler)
        return logger

    def get_types(
        self, request: Optional[InstanceTypeRequest] = None
    ) -> InstanceTypeResponse:
        request = self._prepare_request_with_defaults(request)

        Validator.instance_type_request(request)

        if cached_result := self._cache.get(request):
            self._logger.info("Cache hit")
            return cached_result

        self._logger.info("Cache miss")

        instances = self._fetch_and_describe_instance_types(request)

        filtered_instances = self._apply_filters_to_instances(instances, request)
        sorted_instances = self._sort_instances_by_strategy(filtered_instances, request)

        self._cache_result(request, sorted_instances)
        return sorted_instances

    @staticmethod
    def _prepare_request_with_defaults(
        request: Optional[InstanceTypeRequest],
    ) -> InstanceTypeRequest:
        if request is None:
            request = {}
        request.setdefault("usage_class", UsageClass.ON_DEMAND.value)
        request.setdefault("architecture", Architecture.X86_64.value)
        request.setdefault("product_description", ProductDescription.LINUX_UNIX.value)
        return request

    def _fetch_and_describe_instance_types(
        self, request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        filter_params = self._extract_filter_params(request)
        return self._describe_instance_types(filter_params)

    def _apply_filters_to_instances(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> List[InstanceTypeInfo]:
        self._logger.debug(
            f"Number of instance types before filtering: {len(instances)}"
        )
        filtered_instances = FilterChain(self._region).execute(instances, request)
        self._logger.debug(
            f"Number of instance types after filtering: {len(filtered_instances)}"
        )
        return filtered_instances

    def _sort_instances_by_strategy(
        self, instances: List[InstanceTypeInfo], request: InstanceTypeRequest
    ) -> InstanceTypeResponse:
        strategy = SortStrategyFactory.create(
            request["usage_class"],
            self._region,
            self._pricing_client,
            self._ec2_client,
            self._logger,
            self._spot_price_history_concurrency,
        )
        return strategy.sort(
            instances,
            request["product_description"],
            request.get("availability_zones"),
            request.get("final_spot_price_strategy", "min"),
        )

    def _cache_result(
        self, request: InstanceTypeRequest, result: InstanceTypeResponse
    ) -> None:
        self._cache.set(request, result)

    @staticmethod
    def _extract_filter_params(
        request: InstanceTypeRequest,
    ) -> DescribeInstanceTypeRequest:
        return {
            "is_current_generation": request.get("is_current_generation"),
            "is_instance_storage_supported": request.get(
                "is_instance_storage_supported"
            ),
        }

    @staticmethod
    def _get_client(options: BestEc2Options, service: str, region: str) -> BaseClient:
        if (
            options.get("clients") is not None
            and options["clients"].get(service) is not None
        ):
            return options["clients"][service]
        else:
            return boto3.session.Session().client(service, region_name=region)

    def _describe_instance_types(
        self, options: Optional[DescribeInstanceTypeRequest] = None
    ) -> List[InstanceTypeInfo]:
        filters = []
        if options.get("is_current_generation") is not None:
            is_current_generation = (
                "true" if options.get("is_current_generation") else "false"
            )
            filters.append(
                {
                    "Name": "current-generation",
                    "Values": [is_current_generation],
                }
            )
        if options.get("is_instance_storage_supported") is not None:
            filters.append(
                {
                    "Name": "instance-storage-supported",
                    "Values": [
                        "true" if options["is_instance_storage_supported"] else "false"
                    ],
                }
            )

        instances = []
        paginator = self._ec2_client.get_paginator("describe_instance_types")
        for page in paginator.paginate(Filters=filters):
            instances.extend(page["InstanceTypes"])

        return instances



from .types import InstanceTypeRequest, UsageClass, Architecture


class Validator:
    @staticmethod
    def instance_type_request(request: InstanceTypeRequest) -> None:
        enum_fields = {"usage_class": UsageClass, "architecture": Architecture}
        for field, enum in enum_fields.items():
            if field in request:
                valid_values = {e.value for e in enum}
                if request[field] not in valid_values:
                    raise ValueError(
                        f"Invalid {field} '{request[field]}'. Must be one of {valid_values}."
                    )

        Validator._validate_max_interruption_frequency(request)

    @staticmethod
    def _validate_max_interruption_frequency(request: InstanceTypeRequest) -> None:
        max_interruption_frequency = request.get("max_interruption_frequency")
        if max_interruption_frequency is not None and (
            not isinstance(max_interruption_frequency, int)
            or not (0 <= max_interruption_frequency <= 100)
        ):
            raise ValueError(
                f"Invalid max_interruption_frequency '{max_interruption_frequency}'. "
                "Must be an integer between 9 and 100."
            )



from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional, TypedDict, Union

from botocore.client import BaseClient


class UsageClass(Enum):
    SPOT = "spot"
    ON_DEMAND = "on-demand"


class Architecture(Enum):
    I386 = "i386"
    X86_64 = "x86_64"
    ARM64 = "arm64"
    X86_64_MAC = "x86_64_mac"


class ProductDescription(Enum):
    LINUX_UNIX = "Linux/UNIX"
    RED_HAT_ENTERPRISE_LINUX = "Red Hat Enterprise Linux"
    SUSE_LINUX = "SUSE Linux"
    WINDOWS = "Windows"
    LINUX_UNIX_VPC = "Linux/UNIX (Amazon VPC)"
    RED_HAT_ENTERPRISE_LINUX_VPC = "Red Hat Enterprise Linux (Amazon VPC)"
    SUSE_LINUX_VPC = "SUSE Linux (Amazon VPC)"
    WINDOWS_VPC = "Windows (Amazon VPC)"


class FinalSpotPriceStrategy(Enum):
    MIN = "min"
    MAX = "max"
    AVERAGE = "average"


class InterruptionFrequencyInfo(TypedDict):
    min: int
    max: int
    rate: str


class DiskInfo(TypedDict):
    SizeInGB: int
    Count: int
    Type: str


class InstanceStorageInfo(TypedDict):
    Disks: List[DiskInfo]


class GpuInfo(TypedDict):
    total_gpu_memory: int


class _InstanceTypeRequired(TypedDict):
    price: float
    instance_type: str
    vcpu: int
    memory_gb: int
    network_performance: str
    storage: Union[str, List[DiskInfo]]


# Define the optional fields in a separate class that extends the required one
class InstanceType(_InstanceTypeRequired, total=False):
    az_price: Optional[Dict[str, float]]
    interruption_frequency: Optional[InterruptionFrequencyInfo]
    gpu_memory_gb: Optional[int]


InstanceTypeResponse = List[InstanceType]


class ClientsDict(TypedDict, total=False):
    ec2: Optional[BaseClient]
    pricing: Optional[BaseClient]


class BestEc2Options(TypedDict, total=False):
    region: Optional[str]
    describe_spot_price_history_concurrency: Optional[int]
    describe_on_demand_price_concurrency: Optional[int]
    clients: Optional[ClientsDict]
    cache_ttl_in_minutes: Optional[int]


class NetworkInfo(TypedDict):
    NetworkPerformance: str


class VCpuInfo(TypedDict):
    DefaultVCpus: int


class InstanceTypeInfo(TypedDict):  # total=False makes all keys optional
    InstanceType: str
    CurrentGeneration: bool
    FreeTierEligible: bool
    SupportedUsageClasses: List[str]
    SupportedRootDeviceTypes: List[str]
    SupportedVirtualizationTypes: List[str]
    BareMetal: bool
    Hypervisor: str
    ProcessorInfo: Dict[str, Union[List[str], float, List[str]]]
    VCpuInfo: VCpuInfo
    MemoryInfo: Dict[str, int]
    InstanceStorageSupported: bool
    InstanceStorageInfo: Optional[InstanceStorageInfo]
    EbsInfo: Dict[str, Union[str, Dict[str, Union[int, float]]]]
    NetworkInfo: NetworkInfo
    PlacementGroupInfo: Dict[str, List[str]]
    HibernationSupported: bool
    BurstablePerformanceSupported: bool
    DedicatedHostsSupported: bool
    AutoRecoverySupported: bool
    SupportedBootModes: List[str]
    NitroEnclavesSupport: str
    NitroTpmSupport: str
    NitroTpmInfo: Dict[str, List[str]]
    InterruptionFrequency: Optional[InterruptionFrequencyInfo]
    GpuInfo: Optional[GpuInfo]


class _InstanceTypeRequestRequired(TypedDict):
    vcpu: float
    memory_gb: float


class InstanceTypeRequest(_InstanceTypeRequestRequired, total=False):
    usage_class: Optional[UsageClass]
    burstable: Optional[bool]
    architecture: Optional[Architecture]
    product_description: Optional[ProductDescription]
    is_current_generation: Optional[bool]
    has_gpu: Optional[bool]
    gpu_memory: Optional[int]
    is_instance_storage_supported: Optional[bool]
    max_interruption_frequency: Optional[int]
    availability_zones: Optional[List[str]]
    final_spot_price_strategy: Optional[FinalSpotPriceStrategy]


class ExtendedInstanceTypeRequest(InstanceTypeRequest):
    region: str


class CacheEntry(TypedDict):
    result: InstanceTypeResponse
    datetime: datetime


CacheDict = Dict[str, CacheEntry]


class DescribeInstanceTypeRequest(TypedDict, total=False):
    is_current_generation: Optional[bool]
    is_instance_storage_supported: Optional[bool]


class _PriceDetails(TypedDict):
    price: float


class PriceDetails(_PriceDetails, total=False):
    az_price: Optional[Dict[str, float]]


class TypePriceDetails(TypedDict):
    instance_type: str
    price_details: PriceDetails
